@@ -1,28 +1,33 @@
-use super::util;
+use std::{
+    collections::BTreeMap,
+    ffi::OsStr,
+    fs,
+    io::{self, Write},
+    path::{Path, PathBuf},
+    str,
+};
+
 use ansi_term::Colour;
 use anyhow::{anyhow, Context, Result};
 use difference::{Changeset, Difference};
 use indoc::indoc;
 use lazy_static::lazy_static;
-use regex::bytes::{Regex as ByteRegex, RegexBuilder as ByteRegexBuilder};
-use regex::Regex;
-use std::collections::BTreeMap;
-use std::ffi::OsStr;
-use std::fs;
-use std::io::{self, Write};
-use std::path::{Path, PathBuf};
-use std::str;
+use regex::{
+    bytes::{Regex as ByteRegex, RegexBuilder as ByteRegexBuilder},
+    Regex,
+};
 use tree_sitter::{format_sexp, Language, LogType, Parser, Query};
 use walkdir::WalkDir;
 
+use super::util;
+
 lazy_static! {
     static ref HEADER_REGEX: ByteRegex = ByteRegexBuilder::new(
         r"^(?x)
            (?P<equals>(?:=+){3,})
            (?P<suffix1>[^=\r\n][^\r\n]*)?
            \r?\n
-           (?P<test_name>(?:[^=\r\n:][^\r\n]*\r?\n)+(?:(?:[ \t]*\r?\n)+)?)
-           (?P<markers>((?::(?:skip|error|fail-fast|(language|platform)\([^\r\n)]+\))\r?\n)*))
+           (?P<test_name_and_markers>(?:[^=\r\n][^\r\n]*\r?\n)+)
            ===+
            (?P<suffix2>[^=\r\n][^\r\n]*)?\r?\n"
     )
@@ -301,13 +306,15 @@ fn run_tests(
                             let expected_output = format_sexp(&output, 0);
                             let actual_output = format_sexp(&actual, 0);
 
-                            // Only bail early before updating if the actual is not the output, sometimes
-                            // users want to test cases that are intended to have errors, hence why this
+                            // Only bail early before updating if the actual is not the output,
+                            // sometimes users want to test cases that
+                            // are intended to have errors, hence why this
                             // check isn't shown above
                             if actual.contains("ERROR") || actual.contains("MISSING") {
                                 *has_parse_errors = true;
 
-                                // keep the original `expected` output if the actual output has an error
+                                // keep the original `expected` output if the actual output has an
+                                // error
                                 corrected_entries.push((
                                     name.clone(),
                                     input,
@@ -425,9 +432,9 @@ fn write_tests_to_buffer(
         if i > 0 {
             writeln!(buffer)?;
         }
-        write!(
+        writeln!(
             buffer,
-            "{}\n{name}\n{}\n{input}\n{}\n\n{}\n",
+            "{}\n{name}\n{}\n{input}\n{}\n\n{}",
             "=".repeat(*header_delim_len),
             "=".repeat(*header_delim_len),
             "-".repeat(*divider_delim_len),
@@ -511,29 +518,46 @@ fn parse_test_content(name: String, content: &str, file_path: Option<PathBuf>) -
         let (mut skip, mut platform, mut fail_fast, mut error, mut languages) =
             (false, None, false, false, vec![]);
 
-        let markers = c.name("markers").map_or("".as_bytes(), |m| m.as_bytes());
+        let test_name_and_markers = c
+            .name("test_name_and_markers")
+            .map_or("".as_bytes(), |m| m.as_bytes());
 
-        for marker in markers.split(|&c| c == b'\n').filter(|s| !s.is_empty()) {
-            let marker = str::from_utf8(marker).unwrap();
-            let (marker, right) = marker.split_at(marker.find('(').unwrap_or(marker.len()));
-            match marker {
-                ":skip" => skip = true,
+        let mut test_name = String::new();
+        let mut seen_marker = false;
+
+        for line in str::from_utf8(test_name_and_markers)
+            .unwrap()
+            .lines()
+            .filter(|s| !s.is_empty())
+        {
+            match line.split('(').next().unwrap() {
+                ":skip" => (seen_marker, skip) = (true, true),
                 ":platform" => {
-                    if let Some(platforms) =
-                        right.strip_prefix('(').and_then(|s| s.strip_suffix(')'))
-                    {
+                    if let Some(platforms) = line.strip_prefix(':').and_then(|s| {
+                        s.strip_prefix("platform(")
+                            .and_then(|s| s.strip_suffix(')'))
+                    }) {
+                        seen_marker = true;
                         platform = Some(
                             platform.unwrap_or(false) || platforms.trim() == std::env::consts::OS,
                         );
                     }
                 }
-                ":fail-fast" => fail_fast = true,
-                ":error" => error = true,
+                ":fail-fast" => (seen_marker, fail_fast) = (true, true),
+                ":error" => (seen_marker, error) = (true, true),
                 ":language" => {
-                    if let Some(lang) = right.strip_prefix('(').and_then(|s| s.strip_suffix(')')) {
+                    if let Some(lang) = line.strip_prefix(':').and_then(|s| {
+                        s.strip_prefix("language(")
+                            .and_then(|s| s.strip_suffix(')'))
+                    }) {
+                        seen_marker = true;
                         languages.push(lang.into());
                     }
                 }
+                _ if !seen_marker => {
+                    test_name.push_str(line);
+                    test_name.push('\n');
+                }
                 _ => {}
             }
         }
@@ -550,9 +574,11 @@ fn parse_test_content(name: String, content: &str, file_path: Option<PathBuf>) -
 
         if suffix1 == first_suffix && suffix2 == first_suffix {
             let header_range = c.get(0).unwrap().range();
-            let test_name = c
-                .name("test_name")
-                .map(|c| String::from_utf8_lossy(c.as_bytes()).trim_end().to_string());
+            let test_name = if test_name.is_empty() {
+                None
+            } else {
+                Some(test_name.trim_end().to_string())
+            };
             Some((
                 header_delim_len,
                 header_range,
@@ -636,7 +662,7 @@ fn parse_test_content(name: String, content: &str, file_path: Option<PathBuf>) -
             }
         }
         prev_attributes = attributes;
-        prev_name = test_name.unwrap_or(String::new());
+        prev_name = test_name.unwrap_or_default();
         prev_header_len = header_delim_len;
         prev_header_end = header_range.end;
     }
